import enum
from typing import TYPE_CHECKING, Optional, Union, Tuple, Sequence, Callable

from PyQt6.QtCore import pyqtSignal, Qt, QTimer
from PyQt6.QtGui import QIcon, QPixmap, QColor
from PyQt6.QtWidgets import QLabel, QVBoxLayout, QGridLayout, QPushButton
from PyQt6.QtWidgets import QTreeWidget, QTreeWidgetItem, QHeaderView

from electrum_aionostr.util import from_nip19

from electrum.i18n import _
from electrum.util import NotEnoughFunds, NoDynamicFeeEstimates, UserCancelled, trigger_callback
from electrum.bitcoin import DummyAddress
from electrum.transaction import PartialTxOutput, PartialTransaction, PartialTxInput
from electrum.fee_policy import FeePolicy
from electrum.submarine_swaps import NostrTransport

from electrum.gui.common_qt.util import QtEventListener, qt_event_listener
from electrum.gui import messages

from . import util
from .util import (WindowModalDialog, Buttons, OkButton, CancelButton,
                   EnterButton, ColorScheme, WWLabel, read_QIcon, IconLabel, char_width_in_lineedit,
                   pubkey_to_q_icon)
from .amountedit import BTCAmountEdit
from .fee_slider import FeeSlider, FeeComboBox
from .my_treeview import create_toolbar_with_menu, MyTreeView

if TYPE_CHECKING:
    from .main_window import ElectrumWindow
    from electrum.submarine_swaps import SwapServerTransport, SwapOffer
    from electrum.lnchannel import Channel
    from electrum.simple_config import SimpleConfig

CANNOT_RECEIVE_WARNING = _(
"""The requested amount is higher than what you can receive in your currently open channels.
If you continue, your funds will be locked until the remote server can find a path to pay you.
If the swap cannot be performed after 24h, you will be refunded.
Do you want to continue?"""
)


ROLE_NPUB = Qt.ItemDataRole.UserRole + 1000

class InvalidSwapParameters(Exception): pass


class SwapProvidersButton(QPushButton):

    def __init__(
        self,
        transport_getter: Callable[[], Optional['SwapServerTransport']],
        config: 'SimpleConfig',
        main_window: 'ElectrumWindow',
    ):
        """parent must have a transport() method"""
        QPushButton.__init__(self)
        self.config = config
        self.transport_getter = transport_getter
        self.main_window = main_window
        self.clicked.connect(self.choose_swap_server)
        self.fetching = False
        self.update()

    def update(self):
        if self.fetching:
            self.setEnabled(False)
            self.setText(_("Fetching..."))
            self.setVisible(True)
            return

        transport = self.transport_getter()
        if not isinstance(transport, NostrTransport):
            # HTTPTransport or no Network, not showing server selection button
            self.setEnabled(False)
            self.setVisible(False)
            return
        self.setEnabled(True)
        self.setVisible(True)
        offer_count = len(transport.get_recent_offers())
        button_text = f' {offer_count} ' + (_('swap providers') if offer_count != 1 else _('swap provider'))
        self.setText(button_text)
        # update icon
        if self.config.SWAPSERVER_NPUB:
            pubkey = from_nip19(self.config.SWAPSERVER_NPUB)['object'].hex()
            self.setIcon(pubkey_to_q_icon(pubkey))

    def choose_swap_server(self) -> None:
        transport = self.transport_getter()
        assert isinstance(transport, NostrTransport), transport
        self.main_window.choose_swapserver_dialog(transport)  # type: ignore
        self.update()
        trigger_callback('swap_provider_changed')


class SwapDialog(WindowModalDialog, QtEventListener):
    def __init__(
        self,
        window: 'ElectrumWindow',
        transport: 'SwapServerTransport',
        *,
        is_reverse: Optional[bool] = None,
        recv_amount_sat_or_max: Optional[Union[int, str]] = None,  # sat or '!'
        channels: Optional[Sequence['Channel']] = None,
        get_coins: Optional[Callable[..., Sequence['PartialTxInput']]] = None,
    ):
        WindowModalDialog.__init__(self, window, _('Submarine Swap'))
        self.window = window
        self.config = window.config
        self.lnworker = self.window.wallet.lnworker
        self.swap_manager = self.lnworker.swap_manager
        self.network = window.network
        self.channels = channels
        self.is_reverse = is_reverse if is_reverse is not None else True
        self.get_coins = get_coins

        vbox = QVBoxLayout(self)

        self.transport = transport
        self.server_button = SwapProvidersButton(lambda: self.transport, self.config, self.window)
        self.description_label = WWLabel(self.get_description())
        self.send_amount_e = BTCAmountEdit(self.window.get_decimal_point)
        self.recv_amount_e = BTCAmountEdit(self.window.get_decimal_point)
        self.max_button = EnterButton(_("Max"), self.spend_max)
        btn_width = 10 * char_width_in_lineedit()
        self.max_button.setFixedWidth(btn_width)
        self.max_button.setCheckable(True)
        self.toggle_button = QPushButton('  \U000021c4  ')  # whitespace to force larger min width
        self.toggle_button.setEnabled(is_reverse is None)
        # send_follows is used to know whether the send amount field / receive
        # amount field should be adjusted after the fee slider was moved
        self.send_follows = False
        self.send_amount_e.follows = False
        self.recv_amount_e.follows = False
        self.toggle_button.clicked.connect(self.toggle_direction)
        # textChanged is triggered for both user and automatic action
        self.send_amount_e.textChanged.connect(self.on_send_edited)
        self.recv_amount_e.textChanged.connect(self.on_recv_edited)
        # textEdited is triggered only for user editing of the fields
        self.send_amount_e.textEdited.connect(self.uncheck_max)
        self.recv_amount_e.textEdited.connect(self.uncheck_max)

        self.fee_policy = FeePolicy(self.config.FEE_POLICY)
        self.fee_slider = FeeSlider(parent=self, network=self.network, fee_policy=self.fee_policy, callback=self.fee_slider_callback)
        self.fee_combo = FeeComboBox(self.fee_slider)
        self.fee_target_label = QLabel()
        self._set_fee_slider_visibility(is_visible=not self.is_reverse)

        self.swap_limits_label = QLabel()
        self.fee_label = QLabel()
        self.server_fee_label = QLabel()
        self.last_server_mining_fee_sat = None
        h = QGridLayout()
        h.addWidget(self.description_label, 0, 0, 1, 3)
        h.addWidget(self.toggle_button, 0, 3)
        self.send_label = IconLabel(text=_('You send')+':')
        self.recv_label = IconLabel(text=_('You receive')+':')
        h.addWidget(self.send_label, 1, 0)
        h.addWidget(self.send_amount_e, 1, 1)
        h.addWidget(self.max_button, 1, 2)
        h.addWidget(self.recv_label, 2, 0)
        h.addWidget(self.recv_amount_e, 2, 1)
        h.addWidget(QLabel(_('Swap limits')+':'), 4, 0)
        h.addWidget(self.swap_limits_label, 4, 1, 1, 2)
        h.addWidget(QLabel(_('Server fee')+':'), 5, 0)
        h.addWidget(self.server_fee_label, 5, 1, 1, 2)
        h.addWidget(QLabel(_('Mining fee')+':'), 6, 0)
        h.addWidget(self.fee_label, 6, 1, 1, 2)
        h.addWidget(self.fee_slider, 7, 1)
        h.addWidget(self.fee_combo, 7, 2)
        h.addWidget(self.fee_target_label, 7, 0)
        h.addWidget(QLabel(''), 8, 0)
        vbox.addLayout(h)
        vbox.addStretch()
        self.ok_button = OkButton(self)
        self.ok_button.setDefault(True)
        self.ok_button.setEnabled(False)
        buttons = Buttons(CancelButton(self), self.ok_button)
        vbox.addLayout(buttons)
        buttons.insertWidget(0, self.server_button)
        if recv_amount_sat_or_max:
            assert isinstance(recv_amount_sat_or_max, (int, str)), f"invalid {type(recv_amount_sat_or_max)=}"
            self.init_recv_amount(recv_amount_sat_or_max)
        self.update()
        self.needs_tx_update = True

        self.timer = QTimer(self)
        self.timer.setInterval(500)
        self.timer.setSingleShot(False)
        self.timer.timeout.connect(self.timer_actions)
        self.timer.start()

        self.finished.connect(self.on_finished)

        self.fee_slider.update()
        self.register_callbacks()

    def closeEvent(self, event):
        self.unregister_callbacks()
        event.accept()

    def on_finished(self, *args):
        self.timer.stop()

    @qt_event_listener
    def on_event_fee_histogram(self, *args):
        self.update_send_receive()

    @qt_event_listener
    def on_event_fee(self, *args):
        self.update_send_receive()

    @qt_event_listener
    def on_event_swap_offers_changed(self, recent_offers: Sequence['SwapOffer']):
        self.server_button.update()
        if not self.ok_button.isEnabled():
            # only update the dialog with the new offer if the user hasn't entered an amount yet.
            # if the user has already entered an amount we prefer the swap to fail due to outdated
            # fees than the possibility of a swap happening with fees the user hasn't seen
            # due to an update happening just before the user initiated the swap
            self.update()

    @qt_event_listener
    def on_event_swap_provider_changed(self):
        self.update()
        self.update_send_receive()

    def timer_actions(self):
        if self.needs_tx_update:
            self.update_tx()
            self.update_ok_button()
            self.needs_tx_update = False

    def init_recv_amount(self, recv_amount_sat):
        if recv_amount_sat == '!':
            self.max_button.setChecked(True)
            self.spend_max()
        else:
            recv_amount_sat = max(recv_amount_sat, self.swap_manager.get_min_amount())
            self.recv_amount_e.setAmount(recv_amount_sat)

    def fee_slider_callback(self, fee_rate):
        self.config.FEE_POLICY = self.fee_policy.get_descriptor()
        if not self.is_reverse:
            self.fee_target_label.setText(self.fee_policy.get_target_text())
        self.update_send_receive()
        self.update()

    def _set_fee_slider_visibility(self, *, is_visible: bool):
        if is_visible:
            self.fee_slider.setEnabled(True)
            self.fee_combo.setEnabled(True)
            self.fee_target_label.setText(self.fee_policy.get_target_text())
        else:
            self.fee_slider.setEnabled(False)
            self.fee_combo.setEnabled(False)
            # show the eta of the swap claim
            self.fee_target_label.setText(FeePolicy(self.config.FEE_POLICY_SWAPS).get_target_text())

    def toggle_direction(self):
        self.is_reverse = not self.is_reverse
        self._set_fee_slider_visibility(is_visible=not self.is_reverse)
        self.send_amount_e.setAmount(None)
        self.recv_amount_e.setAmount(None)
        self.max_button.setChecked(False)
        self.update()

    def spend_max(self):
        if self.max_button.isChecked():
            if self.is_reverse:
                self._spend_max_reverse_swap()
            else:
                # spend_max_forward_swap will be called in update_tx
                pass
        else:
            self.send_amount_e.setAmount(None)
        self.needs_tx_update = True

    def uncheck_max(self):
        self.max_button.setChecked(False)
        self.update()

    def _spend_max_forward_swap(self, tx: Optional[PartialTransaction]) -> None:
        if tx:
            amount = tx.output_value_for_address(DummyAddress.SWAP)
            self.send_amount_e.setAmount(amount)
        else:
            self.send_amount_e.setAmount(None)
            self.max_button.setChecked(False)

    def _spend_max_reverse_swap(self) -> None:
        amount = min(self.lnworker.num_sats_can_send(), self.swap_manager.get_provider_max_forward_amount())
        amount = int(amount)  # round down msats
        self.send_amount_e.setAmount(amount)

    def on_send_edited(self):
        if self.send_amount_e.follows:
            return
        self.send_amount_e.setStyleSheet(ColorScheme.DEFAULT.as_stylesheet())
        send_amount = self.send_amount_e.get_amount()
        recv_amount = self.swap_manager.get_recv_amount(send_amount, is_reverse=self.is_reverse)
        if self.is_reverse and send_amount and send_amount > self.lnworker.num_sats_can_send():
            # cannot send this much on lightning
            recv_amount = None
        if (not self.is_reverse) and recv_amount and recv_amount > self.lnworker.num_sats_can_receive():
            # cannot receive this much on lightning
            recv_amount = None
        self.recv_amount_e.follows = True
        self.recv_amount_e.setAmount(recv_amount)
        self.recv_amount_e.setStyleSheet(ColorScheme.BLUE.as_stylesheet())
        self.recv_amount_e.follows = False
        self.send_follows = False
        self.needs_tx_update = True

    def on_recv_edited(self):
        if self.recv_amount_e.follows:
            return
        self.recv_amount_e.setStyleSheet(ColorScheme.DEFAULT.as_stylesheet())
        recv_amount = self.recv_amount_e.get_amount()
        send_amount = self.swap_manager.get_send_amount(recv_amount, is_reverse=self.is_reverse)
        if self.is_reverse and send_amount and send_amount > self.lnworker.num_sats_can_send():
            send_amount = None
        self.send_amount_e.follows = True
        self.send_amount_e.setAmount(send_amount)
        self.send_amount_e.setStyleSheet(ColorScheme.BLUE.as_stylesheet())
        self.send_amount_e.follows = False
        self.send_follows = True
        self.needs_tx_update = True

    def update_send_receive(self):
        self.on_recv_edited() if self.send_follows else self.on_send_edited()

    def update(self):
        sm = self.swap_manager
        w_base_unit = self.window.base_unit()
        send_icon = read_QIcon("lightning.png" if self.is_reverse else "bitcoin.png")
        self.send_label.setIcon(send_icon)
        recv_icon = read_QIcon("lightning.png" if not self.is_reverse else "bitcoin.png")
        self.recv_label.setIcon(recv_icon)
        self.description_label.setText(self.get_description())
        self.description_label.repaint()  # macOS hack for #6269
        min_swap_limit, max_swap_limit = self.get_client_swap_limits_sat()
        if max_swap_limit == 0:
            swap_name = _("reverse") if self.is_reverse else _("forward")
            swap_limit_str = _("No {} swap possible with this provider").format(swap_name)
        else:
            swap_limit_str = (f"{self.window.format_amount(min_swap_limit)} - "
                              f"{self.window.format_amount(max_swap_limit)} {w_base_unit}")
        self.swap_limits_label.setText(swap_limit_str)
        self.swap_limits_label.repaint()  # macOS hack for #6269
        self.last_server_mining_fee_sat = sm.mining_fee
        server_fee_str = '%.2f'%sm.percentage + '%  +  '  + self.window.format_amount(sm.mining_fee) + ' ' + w_base_unit
        self.server_fee_label.setText(server_fee_str)
        self.server_fee_label.repaint()  # macOS hack for #6269
        self.needs_tx_update = True

    def get_client_swap_limits_sat(self) -> Tuple[int, int]:
        """Returns the (min, max) client swap limits in sat."""
        sm = self.swap_manager

        if self.is_reverse:
            lower_limit = sm.get_min_amount()
            upper_limit = sm.client_max_amount_reverse_swap() or 0
        else:
            lower_limit = sm.get_send_amount(sm.get_min_amount(), is_reverse=False) or sm.get_min_amount()
            upper_limit = sm.client_max_amount_forward_swap() or 0

        if lower_limit > upper_limit:
            # if the max possible amount is below the lower limit no swap is possible
            lower_limit, upper_limit = 0, 0
        return lower_limit, upper_limit

    def update_fee(self, tx: Optional[PartialTransaction]) -> None:
        """Updates self.fee_label. No other side-effects."""
        if self.is_reverse:
            sm = self.swap_manager
            fee = sm.get_fee_for_txbatcher()
        else:
            fee = tx.get_fee() if tx else None
        fee_text = self.window.format_amount(fee) + ' ' + self.window.base_unit() if fee else _("no input")
        self.fee_label.setText(fee_text)
        self.fee_label.repaint()  # macOS hack for #6269

    def run(self, transport: 'SwapServerTransport') -> bool:
        """Can raise InvalidSwapParameters."""
        if not self.exec():
            return False
        if self.is_reverse:
            lightning_amount = self.send_amount_e.get_amount()
            onchain_amount = self.recv_amount_e.get_amount()
            if lightning_amount is None or onchain_amount is None:
                return False
            sm = self.swap_manager
            coro = sm.reverse_swap(
                transport=transport,
                lightning_amount_sat=lightning_amount,
                expected_onchain_amount_sat=onchain_amount + self.swap_manager.get_fee_for_txbatcher(),
                prepayment_sat=2 * self.last_server_mining_fee_sat,
            )
            try:
                # we must not leave the context, so we use run_couroutine_dialog
                funding_txid = self.window.run_coroutine_dialog(coro, _('Initiating swap...'))
            except Exception as e:
                self.window.show_error(f"Reverse swap failed: {str(e)}")
                return False
            self.window.on_swap_result(funding_txid, is_reverse=True)
            return True
        else:
            lightning_amount = self.recv_amount_e.get_amount()
            onchain_amount = self.send_amount_e.get_amount()
            if lightning_amount is None or onchain_amount is None:
                return False
            if lightning_amount > self.lnworker.num_sats_can_receive():
                if not self.window.question(CANNOT_RECEIVE_WARNING):
                    return False
            self.window.protect(self.do_normal_swap, (transport, lightning_amount, onchain_amount))
            return True

    def update_tx(self) -> None:
        if self.is_reverse:
            self.update_fee(None)
            return
        is_max = self.max_button.isChecked()
        if is_max:
            tx = self._create_tx_safe('!')
            self._spend_max_forward_swap(tx)
        else:
            onchain_amount = self.send_amount_e.get_amount()
            tx = self._create_tx_safe(onchain_amount)
        self.update_fee(tx)

    def _create_tx(self, onchain_amount: Union[int, str, None]) -> PartialTransaction:
        assert not self.is_reverse
        if onchain_amount is None:
            raise InvalidSwapParameters("onchain_amount is None")
        coins = self.get_coins() if self.get_coins else self.window.get_coins()
        if onchain_amount == '!':
            max_amount = sum(c.value_sats() for c in coins)
            max_swap_amount = self.swap_manager.client_max_amount_forward_swap()
            if max_swap_amount is None:
                raise InvalidSwapParameters("swap_manager.client_max_amount_forward_swap() is None")
            if max_amount > max_swap_amount:
                onchain_amount = max_swap_amount
        outputs = [PartialTxOutput.from_address_and_value(DummyAddress.SWAP, onchain_amount)]
        try:
            tx = self.window.wallet.make_unsigned_transaction(
                fee_policy=self.fee_policy,
                coins=coins,
                outputs=outputs,
                send_change_to_lightning=False,
            )
        except (NotEnoughFunds, NoDynamicFeeEstimates) as e:
            raise InvalidSwapParameters(str(e)) from e
        return tx

    def _create_tx_safe(self, onchain_amount: Union[int, str, None]) -> Optional[PartialTransaction]:
        try:
            return self._create_tx(onchain_amount=onchain_amount)
        except InvalidSwapParameters:
            return None

    def update_ok_button(self):
        """Updates self.ok_button. No other side-effects."""
        send_amount = self.send_amount_e.get_amount()
        recv_amount = self.recv_amount_e.get_amount()
        self.ok_button.setEnabled(bool(send_amount) and bool(recv_amount))

    async def _do_normal_swap(self, transport, lightning_amount, onchain_amount, password):
        dummy_tx = self._create_tx(onchain_amount)
        assert dummy_tx
        sm = self.swap_manager
        swap, invoice = await sm.request_normal_swap(
            transport=transport,
            lightning_amount_sat=lightning_amount,
            expected_onchain_amount_sat=onchain_amount,
            channels=self.channels,
        )
        self._current_swap = swap
        tx = sm.create_funding_tx(swap, dummy_tx, password=password)
        txid = await sm.wait_for_htlcs_and_broadcast(transport=transport, swap=swap, invoice=invoice, tx=tx)
        return txid

    def do_normal_swap(self, transport, lightning_amount, onchain_amount, password):
        self._current_swap = None
        coro = self._do_normal_swap(transport, lightning_amount, onchain_amount, password)
        try:
            funding_txid = self.window.run_coroutine_dialog(coro, _('Awaiting swap payment...'))
        except UserCancelled:
            if self._current_swap is None or self.swap_manager.cancel_normal_swap(self._current_swap):
                self.window.show_message(_('Swap cancelled'))
            else:
                self.window.show_message(_('Swap was already funded, refusing to cancel'))
            return
        except Exception as e:
            self.window.show_error(str(e))
            return
        self.window.on_swap_result(funding_txid, is_reverse=False)

    def get_description(self):
        onchain_funds = "onchain"
        lightning_funds = "lightning"

        return "Send {fromType}, receive {toType}.\nThis will increase your lightning {capacityType} capacity.\n".format(
            fromType=lightning_funds if self.is_reverse else onchain_funds,
            toType=onchain_funds if self.is_reverse else lightning_funds,
            capacityType="receiving" if self.is_reverse else "sending",
        )


class SwapServerDialog(WindowModalDialog, QtEventListener):

    class Columns(MyTreeView.BaseColumnsEnum):
        PUBKEY = enum.auto()
        FEE = enum.auto()
        MAX_FORWARD = enum.auto()
        MAX_REVERSE = enum.auto()
        LAST_SEEN = enum.auto()

    headers = {
        Columns.PUBKEY: _("Pubkey"),
        Columns.FEE: _("Fee"),
        Columns.MAX_FORWARD: _('Max Forward'),
        Columns.MAX_REVERSE: _('Max Reverse'),
        Columns.LAST_SEEN: _("Last seen"),
    }

    def __init__(self, window: 'ElectrumWindow', servers: Sequence['SwapOffer']):
        WindowModalDialog.__init__(self, window, _('Choose Swap Provider'))
        self.window = window
        self.config = window.config
        msg = '\n'.join([
            _("Please choose a provider from this list."),
            _("Note that fees and liquidity may be updated frequently.")
        ])
        self.servers_list = QTreeWidget()
        col_names = [self.headers[col_idx] for col_idx in sorted(self.headers.keys())]
        self.servers_list.setHeaderLabels(col_names)
        self.servers_list.header().setStretchLastSection(False)
        for col_idx in range(len(self.Columns)):
            sm = QHeaderView.ResizeMode.Stretch if col_idx == self.Columns.PUBKEY else QHeaderView.ResizeMode.ResizeToContents
            self.servers_list.header().setSectionResizeMode(col_idx, sm)
        self.update_servers_list(servers)
        vbox = QVBoxLayout()
        self.setLayout(vbox)
        vbox.addWidget(WWLabel(msg))
        vbox.addWidget(self.servers_list, stretch=1)
        vbox.addSpacing(10)
        self.ok_button = OkButton(self)
        vbox.addLayout(Buttons(CancelButton(self), self.ok_button))
        self.setMinimumWidth(650)
        self.register_callbacks()

    def run(self):
        if self.exec() != 1:
            return None
        if item := self.servers_list.currentItem():
            return item.data(self.Columns.PUBKEY, ROLE_NPUB)
        return None

    def closeEvent(self, event):
        self.unregister_callbacks()
        event.accept()

    @qt_event_listener
    def on_event_swap_offers_changed(self, recent_offers: Sequence['SwapOffer']):
        self.update_servers_list(recent_offers)

    def update_servers_list(self, servers: Sequence['SwapOffer']):
        self.servers_list.clear()
        from electrum.util import age
        items = []
        for x in servers:
            labels = [""] * len(self.Columns)
            labels[self.Columns.PUBKEY] = x.server_pubkey
            labels[self.Columns.FEE] = f"{x.pairs.percentage}% + {x.pairs.mining_fee} sats"
            labels[self.Columns.MAX_FORWARD] = self.window.format_amount(x.pairs.max_forward) + ' ' + self.window.base_unit()
            labels[self.Columns.MAX_REVERSE] = self.window.format_amount(x.pairs.max_reverse) + ' ' + self.window.base_unit()
            labels[self.Columns.LAST_SEEN] = age(x.timestamp)
            item = QTreeWidgetItem(labels)
            item.setData(self.Columns.PUBKEY, ROLE_NPUB, x.server_npub)
            item.setIcon(self.Columns.PUBKEY, pubkey_to_q_icon(x.server_pubkey))
            items.append(item)
        self.servers_list.insertTopLevelItems(0, items)




#!/usr/bin/env python
#
# Electrum - lightweight Bitcoin client
# Copyright (C) 2025 The Electrum Developers
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import asyncio
import hashlib
import json
from typing import Union, TYPE_CHECKING

import base64

from electrum import util
from electrum.plugin import BasePlugin, hook
from electrum.crypto import aes_encrypt_with_iv, aes_decrypt_with_iv
from electrum.i18n import _
from electrum.util import log_exceptions, ignore_exceptions, make_aiohttp_session
from electrum.network import Network

if TYPE_CHECKING:
    from electrum.wallet import Abstract_Wallet


class ErrorConnectingServer(Exception):
    def __init__(self, reason: Union[str, Exception] = None):
        self.reason = reason

    def __str__(self):
        header = _("Error connecting to {} server").format('Labels')
        reason = self.reason
        if isinstance(reason, BaseException):
            reason = repr(reason)
        return f"{header}: {reason}" if reason else header


class LabelsPlugin(BasePlugin):

    def __init__(self, parent, config, name):
        BasePlugin.__init__(self, parent, config, name)
        self.target_host = 'labels.electrum.org'
        self.wallets = {}

    def encode(self, wallet: 'Abstract_Wallet', msg: str) -> str:
        password, iv, wallet_id = self.wallets[wallet]
        encrypted = aes_encrypt_with_iv(password, iv, msg.encode('utf8'))
        # FIXME: ^ we are reusing the IV between all labels in the wallet, in CBC mode...
        return base64.b64encode(encrypted).decode()

    def decode(self, wallet: 'Abstract_Wallet', message: str) -> str:
        password, iv, wallet_id = self.wallets[wallet]
        decoded = base64.b64decode(message, validate=True)
        decrypted = aes_decrypt_with_iv(password, iv, decoded)
        return decrypted.decode('utf8')

    def get_nonce(self, wallet: 'Abstract_Wallet'):
        # nonce is the nonce to be used with the next change
        nonce = wallet.db.get('wallet_nonce')
        if nonce is None:
            nonce = 1
            self.set_nonce(wallet, nonce)
        return nonce

    def set_nonce(self, wallet: 'Abstract_Wallet', nonce):
        self.logger.info(f"set {wallet.basename()} nonce to {nonce}")
        wallet.db.put("wallet_nonce", nonce)

    @hook
    def set_label(self, wallet: 'Abstract_Wallet', item, label):
        if wallet not in self.wallets:
            return
        if not item:
            return
        if label is None:
            # note: the server should not know whether a label is empty
            #       FIXME but it does! we are reusing the IV with AES-CBC: there is no randomness between labels,
            #       all empty labels in given wallet look the same.
            label = ''
        nonce = self.get_nonce(wallet)
        wallet_id = self.wallets[wallet][2]
        bundle = {
            "walletId": wallet_id,
            "walletNonce": nonce,
            "externalId": self.encode(wallet, item),
            "encryptedLabel": self.encode(wallet, label)
        }
        asyncio.run_coroutine_threadsafe(self.do_post_safe("/label", bundle), wallet.network.asyncio_loop)
        # Caller will write the wallet
        self.set_nonce(wallet, nonce + 1)

    @ignore_exceptions
    @log_exceptions
    async def do_post_safe(self, *args):
        await self.do_post(*args)

    async def do_get(self, url = "/labels"):
        url = 'https://' + self.target_host + url
        network = Network.get_instance()
        proxy = network.proxy if network else None
        async with make_aiohttp_session(proxy) as session:
            async with session.get(url) as result:
                return await result.json()

    async def do_post(self, url = "/labels", data=None):
        url = 'https://' + self.target_host + url
        network = Network.get_instance()
        proxy = network.proxy if network else None
        async with make_aiohttp_session(proxy) as session:
            async with session.post(url, json=data) as result:
                try:
                    return await result.json()
                except Exception as e:
                    raise Exception('Could not decode: ' + await result.text()) from e

    async def push_thread(self, wallet: 'Abstract_Wallet') -> int:
        wallet_data = self.wallets.get(wallet, None)
        if not wallet_data:
            raise Exception('Wallet {} not loaded'.format(wallet))
        wallet_id = wallet_data[2]
        bundle = {"labels": [],
                  "walletId": wallet_id,
                  "walletNonce": self.get_nonce(wallet)}
        for key, value in wallet.get_all_labels().items():
            try:
                encoded_key = self.encode(wallet, key)
                encoded_value = self.encode(wallet, value)
            except Exception:
                self.logger.info(f'cannot encode {repr(key)} {repr(value)}')
                continue
            bundle["labels"].append({'encryptedLabel': encoded_value,
                                     'externalId': encoded_key})
        await self.do_post("/labels", bundle)
        return len(bundle['labels'])

    async def pull_thread(self, wallet: 'Abstract_Wallet', force: bool) -> int:
        wallet_data = self.wallets.get(wallet, None)
        if not wallet_data:
            raise Exception('Wallet {} not loaded'.format(wallet))
        wallet_id = wallet_data[2]
        nonce = 1 if force else self.get_nonce(wallet) - 1
        self.logger.info(f"asking for labels since nonce {nonce}")
        try:
            response = await self.do_get("/labels/since/%d/for/%s" % (nonce, wallet_id))
        except Exception as e:
            raise ErrorConnectingServer(e) from e
        if response["labels"] is None or len(response["labels"]) == 0:
            self.logger.info('no new labels')
            return 0

        self.logger.info(f'received {len(response["labels"])} labels')
        result = {}
        for label in response["labels"]:
            try:
                key = self.decode(wallet, label["externalId"])
                value = self.decode(wallet, label["encryptedLabel"])
            except Exception:
                continue
            try:
                json.dumps(key)
                json.dumps(value)
            except Exception:
                self.logger.info(f'error: no json {key}')
                continue
            if value:
                result[key] = value

        for key, value in result.items():
            wallet._set_label(key, value)

        self.set_nonce(wallet, response["nonce"] + 1)
        util.trigger_callback('labels_received', wallet, result)
        self.on_pulled(wallet)
        return len(result)

    def on_pulled(self, wallet: 'Abstract_Wallet') -> None:
        pass

    @ignore_exceptions
    @log_exceptions
    async def pull_safe_thread(self, wallet: 'Abstract_Wallet', force: bool):
        try:
            await self.pull_thread(wallet, force)
        except ErrorConnectingServer as e:
            self.logger.info(repr(e))

    def pull(self, wallet: 'Abstract_Wallet', force: bool):
        if not wallet.network:
            raise Exception(_('You are offline.'))
        return asyncio.run_coroutine_threadsafe(self.pull_thread(wallet, force), wallet.network.asyncio_loop).result()

    def push(self, wallet: 'Abstract_Wallet'):
        if not wallet.network:
            raise Exception(_('You are offline.'))
        return asyncio.run_coroutine_threadsafe(self.push_thread(wallet), wallet.network.asyncio_loop).result()

    def start_wallet(self, wallet: 'Abstract_Wallet'):
        """Labels have the same level of privacy as the wallet transaction
        history. Since the wallet master public key(s) give access to
        the transaction history, we also use it to encrypt labels.
        """
        if not wallet.network:
            return  # 'offline' mode
        mpk = wallet.get_fingerprint()
        if not mpk:
            return
        mpk = mpk.encode('ascii')
        password = hashlib.sha1(mpk).hexdigest()[:32].encode('ascii')
        iv = hashlib.sha256(password).digest()[:16]
        wallet_id = hashlib.sha256(mpk).hexdigest()
        self.wallets[wallet] = (password, iv, wallet_id)
        nonce = self.get_nonce(wallet)
        self.logger.info(f"wallet {wallet.basename()} nonce is {nonce}")
        # If there is an auth token we can try to actually start syncing
        asyncio.run_coroutine_threadsafe(self.pull_safe_thread(wallet, False), wallet.network.asyncio_loop)

    def stop_wallet(self, wallet):
        self.wallets.pop(wallet, None)



#!/usr/bin/env python
#
# Electrum - lightweight Bitcoin client
# Copyright (C) 2012 thomasv@gitorious
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from typing import TYPE_CHECKING

from PyQt6.QtWidgets import QVBoxLayout, QLabel

from electrum.i18n import _

from .util import WindowModalDialog, ButtonsLineEdit, ShowQRLineEdit, Buttons, CloseButton
from .history_list import HistoryList, HistoryModel
from .qrtextedit import ShowQRTextEdit

if TYPE_CHECKING:
    from .main_window import ElectrumWindow


class AddressHistoryModel(HistoryModel):
    def __init__(self, window: 'ElectrumWindow', address):
        super().__init__(window)
        self.address = address

    def get_domain(self):
        return [self.address]

    def should_include_lightning_payments(self) -> bool:
        return False


class AddressDialog(WindowModalDialog):

    def __init__(self, window: 'ElectrumWindow', address: str, *, parent=None):
        if parent is None:
            parent = window
        WindowModalDialog.__init__(self, parent, _("Address"))
        self.address = address
        self.window = window
        self.config = window.config
        self.wallet = window.wallet
        self.app = window.app
        self.saved = True

        self.setMinimumWidth(700)
        vbox = QVBoxLayout()
        self.setLayout(vbox)

        vbox.addWidget(QLabel(_("Address") + ":"))
        self.addr_e = ShowQRLineEdit(self.address, self.config, title=_("Address"))
        vbox.addWidget(self.addr_e)

        try:
            pubkeys = self.wallet.get_public_keys(address)
        except BaseException as e:
            pubkeys = None
        if pubkeys:
            vbox.addWidget(QLabel(_("Public keys") + ':'))
            for pubkey in pubkeys:
                pubkey_e = ShowQRLineEdit(pubkey, self.config, title=_("Public Key"))
                vbox.addWidget(pubkey_e)

        redeem_script = self.wallet.get_redeem_script(address)
        if redeem_script:
            vbox.addWidget(QLabel(_("Redeem Script") + ':'))
            redeem_e = ShowQRTextEdit(text=redeem_script, config=self.config)
            redeem_e.addCopyButton()
            vbox.addWidget(redeem_e)

        witness_script = self.wallet.get_witness_script(address)
        if witness_script:
            vbox.addWidget(QLabel(_("Witness Script") + ':'))
            witness_e = ShowQRTextEdit(text=witness_script, config=self.config)
            witness_e.addCopyButton()
            vbox.addWidget(witness_e)

        address_path_str = self.wallet.get_address_path_str(address)
        if address_path_str:
            vbox.addWidget(QLabel(_("Derivation path") + ':'))
            der_path_e = ButtonsLineEdit(address_path_str)
            der_path_e.addCopyButton()
            der_path_e.setReadOnly(True)
            vbox.addWidget(der_path_e)

        addr_hist_model = AddressHistoryModel(self.window, self.address)
        self.hw = HistoryList(self.window, addr_hist_model)
        self.hw.num_tx_label = QLabel('')
        addr_hist_model.set_view(self.hw)
        vbox.addWidget(self.hw.num_tx_label)
        vbox.addWidget(self.hw)

        vbox.addLayout(Buttons(CloseButton(self)))
        self.format_amount = self.window.format_amount
        addr_hist_model.refresh('address dialog constructor')

    def show_qr(self):
        text = self.address
        try:
            self.window.show_qrcode(text, 'Address', parent=self)
        except Exception as e:
            self.show_message(repr(e))



#!/usr/bin/env python
#
# Electrum - lightweight Bitcoin client
# Copyright (C) 2013 ecdsa@github
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from typing import TYPE_CHECKING

from PyQt6.QtWidgets import QVBoxLayout, QHBoxLayout, QLabel, QWidget, QGridLayout, QToolButton, QPushButton
from PyQt6.QtCore import QRect, Qt
from PyQt6.QtGui import QPen, QPainter, QPixmap

from electrum.i18n import _
from electrum.gui.messages import MSG_LN_UTXO_RESERVE

from .util import Buttons, CloseButton, WindowModalDialog, ColorScheme, font_height, AmountLabel, icon_path

if TYPE_CHECKING:
    from .main_window import ElectrumWindow
    from electrum.wallet import Abstract_Wallet


# Todo:
#  show lightning funds that are not usable
#  pie chart mouse interactive, to prepare a swap

COLOR_CONFIRMED = Qt.GlobalColor.green
COLOR_UNCONFIRMED = Qt.GlobalColor.red
COLOR_UNMATURED = Qt.GlobalColor.magenta
COLOR_FROZEN = ColorScheme.BLUE.as_color(True)
COLOR_LIGHTNING = Qt.GlobalColor.yellow
COLOR_FROZEN_LIGHTNING = Qt.GlobalColor.cyan


class PieChartObject:

    def paintEvent(self, event):
        pen = QPen(Qt.GlobalColor.gray, 1, Qt.PenStyle.SolidLine)
        qp = QPainter()
        qp.begin(self)
        qp.setPen(pen)
        qp.setRenderHint(QPainter.RenderHint.Antialiasing)
        qp.setBrush(Qt.GlobalColor.gray)
        total = sum([x[2] for x in self._list])
        if total == 0:
            return
        alpha = 0
        s = 0
        for name, color, amount in self._list:
            qp.setBrush(color)
            if amount == 0:
                continue
            elif amount == total:
                qp.drawEllipse(self.R)
            else:
                delta = int(16 * 360 * amount/total)
                qp.drawPie(self.R, alpha, delta)
                alpha += delta
        qp.end()


class PieChartWidget(QWidget, PieChartObject):

    def __init__(self, size, l):
        QWidget.__init__(self)
        self.size = size
        self.R = QRect(0, 0, self.size, self.size)
        self.setGeometry(self.R)
        self.setMinimumWidth(self.size)
        self.setMaximumWidth(self.size)
        self.setMinimumHeight(self.size)
        self.setMaximumHeight(self.size)
        self._list = l # list[ (name, color, amount)]
        self.update()

    def update_list(self, l):
        self._list = l
        self.update()


class BalanceToolButton(QToolButton, PieChartObject):

    def __init__(self):
        QToolButton.__init__(self)
        self._list = []
        self._update_size()
        self._warning = False

    @property
    def has_warning(self) -> bool:
        return bool(self._warning)

    def update_list(self, l, warning: bool):
        self._warning = warning
        self._list = l
        self.update()

    def setText(self, text):
        # this is a hack
        QToolButton.setText(self, '       ' + text)

    def paintEvent(self, event):
        QToolButton.paintEvent(self, event)
        if not self._warning:
            PieChartObject.paintEvent(self, event)
        else:
            pixmap = QPixmap(icon_path("warning.png"))
            qp = QPainter()
            qp.begin(self)
            qp.drawPixmap(self.R, pixmap)
            qp.end()

    def resizeEvent(self, e):
        super().resizeEvent(e)
        self._update_size()

    def _update_size(self):
        size = round(font_height(self) * 1.1)
        self.R = QRect(6, 3, size, size)


class LegendWidget(QWidget):
    size = 20

    def __init__(self, color):
        QWidget.__init__(self)
        self.color = color
        self.R = QRect(0, 0, self.size, int(self.size*0.75))
        self.setGeometry(self.R)
        self.setMinimumWidth(self.size)
        self.setMaximumWidth(self.size)
        self.setMinimumHeight(self.size)
        self.setMaximumHeight(self.size)

    def paintEvent(self, event):
        pen = QPen(Qt.GlobalColor.gray, 1, Qt.PenStyle.SolidLine)
        qp = QPainter()
        qp.begin(self)
        qp.setPen(pen)
        qp.setRenderHint(QPainter.RenderHint.Antialiasing)
        qp.setBrush(self.color)
        qp.drawRect(self.R)
        qp.end()


class BalanceDialog(WindowModalDialog):

    def __init__(self, parent: 'ElectrumWindow', *, wallet: 'Abstract_Wallet'):

        WindowModalDialog.__init__(self, parent, _("Wallet Balance"))
        self.wallet = wallet
        self.window = parent
        self.config = parent.config
        self.fx = parent.fx

        p_bal = self.wallet.get_balances_for_piechart()
        confirmed = p_bal.confirmed
        unconfirmed = p_bal.unconfirmed
        unmatured = p_bal.unmatured
        frozen = p_bal.frozen
        lightning = p_bal.lightning
        f_lightning = p_bal.lightning_frozen

        frozen_str =  self.config.format_amount_and_units(frozen)
        confirmed_str =  self.config.format_amount_and_units(confirmed)
        unconfirmed_str =  self.config.format_amount_and_units(unconfirmed)
        unmatured_str =  self.config.format_amount_and_units(unmatured)
        lightning_str =  self.config.format_amount_and_units(lightning)
        f_lightning_str =  self.config.format_amount_and_units(f_lightning)

        frozen_fiat_str = self.fx.format_amount_and_units(frozen) if self.fx else ''
        confirmed_fiat_str = self.fx.format_amount_and_units(confirmed) if self.fx else ''
        unconfirmed_fiat_str = self.fx.format_amount_and_units(unconfirmed) if self.fx else ''
        unmatured_fiat_str = self.fx.format_amount_and_units(unmatured) if self.fx else ''
        lightning_fiat_str = self.fx.format_amount_and_units(lightning) if self.fx else ''
        f_lightning_fiat_str = self.fx.format_amount_and_units(f_lightning) if self.fx else ''

        piechart = PieChartWidget(
            max(120, 9 * font_height()),
            [
                (_('Frozen'), COLOR_FROZEN, frozen),
                (_('Unmatured'), COLOR_UNMATURED, unmatured),
                (_('Unconfirmed'), COLOR_UNCONFIRMED, unconfirmed),
                (_('On-chain'), COLOR_CONFIRMED, confirmed),
                (_('Lightning'), COLOR_LIGHTNING, lightning),
                (_('Lightning frozen'), COLOR_FROZEN_LIGHTNING, f_lightning),
            ]
        )

        vbox = QVBoxLayout()
        if self.wallet.is_low_reserve():
            reserve_str = self.config.format_amount_and_units(self.config.LN_UTXO_RESERVE)
            hbox = QHBoxLayout()
            msg = _('Warning') + ': ' + MSG_LN_UTXO_RESERVE.format(reserve_str)
            label = QLabel(msg)
            label.setWordWrap(True)
            logo = QLabel('')
            logo.setPixmap(
                QPixmap(icon_path("warning.png")).scaledToWidth(
                    25, mode=Qt.TransformationMode.SmoothTransformation)
            )
            logo.setMaximumWidth(28)
            hbox.addWidget(logo)
            hbox.addWidget(label)
            vbox.addLayout(hbox)

        vbox.addWidget(piechart)
        grid = QGridLayout()
        #grid.addWidget(QLabel(_("Onchain") + ':'), 0, 1)
        #grid.addWidget(QLabel(onchain_str), 0, 2, alignment=Qt.AlignmentFlag.AlignRight)
        #grid.addWidget(QLabel(onchain_fiat_str), 0, 3, alignment=Qt.AlignmentFlag.AlignRight)

        if frozen:
            grid.addWidget(LegendWidget(COLOR_FROZEN), 0, 0)
            grid.addWidget(QLabel(_("Frozen") + ':'), 0, 1)
            grid.addWidget(AmountLabel(frozen_str), 0, 2, alignment=Qt.AlignmentFlag.AlignRight)
            grid.addWidget(AmountLabel(frozen_fiat_str), 0, 3, alignment=Qt.AlignmentFlag.AlignRight)
        if unconfirmed:
            grid.addWidget(LegendWidget(COLOR_UNCONFIRMED), 2, 0)
            grid.addWidget(QLabel(_("Unconfirmed") + ':'), 2, 1)
            grid.addWidget(AmountLabel(unconfirmed_str), 2, 2, alignment=Qt.AlignmentFlag.AlignRight)
            grid.addWidget(AmountLabel(unconfirmed_fiat_str), 2, 3, alignment=Qt.AlignmentFlag.AlignRight)
        if unmatured:
            grid.addWidget(LegendWidget(COLOR_UNMATURED), 3, 0)
            grid.addWidget(QLabel(_("Unmatured") + ':'), 3, 1)
            grid.addWidget(AmountLabel(unmatured_str), 3, 2, alignment=Qt.AlignmentFlag.AlignRight)
            grid.addWidget(AmountLabel(unmatured_fiat_str), 3, 3, alignment=Qt.AlignmentFlag.AlignRight)
        if confirmed:
            grid.addWidget(LegendWidget(COLOR_CONFIRMED), 1, 0)
            grid.addWidget(QLabel(_("On-chain") + ':'), 1, 1)
            grid.addWidget(AmountLabel(confirmed_str), 1, 2, alignment=Qt.AlignmentFlag.AlignRight)
            grid.addWidget(AmountLabel(confirmed_fiat_str), 1, 3, alignment=Qt.AlignmentFlag.AlignRight)
        if lightning:
            grid.addWidget(LegendWidget(COLOR_LIGHTNING), 4, 0)
            grid.addWidget(QLabel(_("Lightning") + ':'), 4, 1)
            grid.addWidget(AmountLabel(lightning_str), 4, 2, alignment=Qt.AlignmentFlag.AlignRight)
            grid.addWidget(AmountLabel(lightning_fiat_str), 4, 3, alignment=Qt.AlignmentFlag.AlignRight)
        if f_lightning:
            grid.addWidget(LegendWidget(COLOR_FROZEN_LIGHTNING), 5, 0)
            grid.addWidget(QLabel(_("Lightning (frozen)") + ':'), 5, 1)
            grid.addWidget(AmountLabel(f_lightning_str), 5, 2, alignment=Qt.AlignmentFlag.AlignRight)
            grid.addWidget(AmountLabel(f_lightning_fiat_str), 5, 3, alignment=Qt.AlignmentFlag.AlignRight)

        vbox.addLayout(grid)
        vbox.addStretch(1)
        buttons = [CloseButton(self)]
        if self.window.wallet.has_lightning():
            swap_button = QPushButton(_('Swap'))
            swap_button.clicked.connect(lambda: self.window.run_swap_dialog())
            buttons.insert(0, swap_button)

        vbox.addLayout(Buttons(*buttons))
        self.setLayout(vbox)

    def run(self):
        self.exec()



# Copyright (C) 2019 The Electrum developers
# Distributed under the MIT software license, see the accompanying
# file LICENCE or http://www.opensource.org/licenses/mit-license.php

import logging
import logging.handlers
import datetime
import sys
import pathlib
import os
import platform
from typing import Optional, TYPE_CHECKING, Set
import copy
import subprocess
import hashlib

if TYPE_CHECKING:
    from .simple_config import SimpleConfig


class LogFormatterForFiles(logging.Formatter):

    def formatTime(self, record, datefmt=None):
        # timestamps follow ISO 8601 UTC
        date = datetime.datetime.fromtimestamp(record.created).astimezone(datetime.timezone.utc)
        if not datefmt:
            datefmt = "%Y%m%dT%H%M%S.%fZ"
        return date.strftime(datefmt)

    def format(self, record):
        record = _shorten_name_of_logrecord(record)
        return super().format(record)


file_formatter = LogFormatterForFiles(fmt="%(asctime)22s | %(levelname)8s | %(name)s | %(message)s")


class LogFormatterForConsole(logging.Formatter):

    def formatTime(self, record, datefmt=None):
        t = record.relativeCreated / 1000
        return f"{t:6.2f}"

    def format(self, record):
        record = copy.copy(record)  # avoid mutating arg
        record = _shorten_name_of_logrecord(record)
        text = super().format(record)
        return text


# try to make console log lines short... no timestamp, short levelname, no "electrum."
console_formatter = LogFormatterForConsole(fmt="%(asctime)s | %(levelname).1s | %(name)s | %(message)s")


def _shorten_name_of_logrecord(record: logging.LogRecord) -> logging.LogRecord:
    record = copy.copy(record)  # avoid mutating arg
    # strip the main module name from the logger name
    if record.name.startswith("electrum."):
        record.name = record.name[9:]
    # manual map to shorten common module names
    record.name = record.name.replace("interface.Interface", "interface", 1)
    record.name = record.name.replace("network.Network", "network", 1)
    record.name = record.name.replace("synchronizer.Synchronizer", "synchronizer", 1)
    record.name = record.name.replace("verifier.SPV", "verifier", 1)
    record.name = record.name.replace("gui.qt.main_window.ElectrumWindow", "gui.qt.main_window", 1)
    return record


class TruncatingMemoryHandler(logging.handlers.MemoryHandler):
    """An in-memory log handler that only keeps the first N log messages
    and discards the rest.
    """
    target: Optional['logging.Handler']

    def __init__(self):
        logging.handlers.MemoryHandler.__init__(
            self,
            capacity=1,  # note: this is the flushing frequency, ~unused by us
            flushLevel=logging.DEBUG,
        )
        self.max_size = 100  # max num of messages we keep
        self.num_messages_seen = 0
        self.__never_dumped = True

    # note: this flush implementation *keeps* the buffer as-is, instead of clearing it
    def flush(self):
        self.acquire()
        try:
            if self.target:
                for record in self.buffer:
                    if record.levelno >= self.target.level:
                        self.target.handle(record)
        finally:
            self.release()

    def dump_to_target(self, target: 'logging.Handler'):
        self.acquire()
        try:
            self.setTarget(target)
            self.flush()
            self.setTarget(None)
        finally:
            self.__never_dumped = False
            self.release()

    def emit(self, record):
        self.num_messages_seen += 1
        if len(self.buffer) < self.max_size:
            super().emit(record)

    def close(self) -> None:
        # Check if captured log lines were never to dumped to e.g. stderr,
        # and if so, try to do it now. This is useful e.g. in case of sys.exit().
        if self.__never_dumped:
            _configure_stderr_logging()
        super().close()


def _delete_old_logs(path, *, num_files_keep: int, max_total_size: int):
    """Delete old logfiles, only keeping the latest few."""
    def sortkey_oldest_first(p: pathlib.PurePath):
        fname = p.name
        basename, ext, counter = str(fname).partition(".log")
        # - each time electrum is launched, there will be a new basename, ordered by date
        # - for any given basename, there might be multiple log files, differing by counter
        #   - empty counter is newest, then .1 is older, .2 is even older, etc
        try:
            counter = int(counter[1:]) if counter else 0  # convert ".2" -> 2
        except ValueError:
            _logger.warning(f"failed to parse log file name: {fname}")
            counter = 0
        return basename, -counter
    files = sorted(
        list(pathlib.Path(path).glob("electrum_log_*.log*")),
        key=sortkey_oldest_first,
    )
    total_size = sum(os.stat(f).st_size for f in files)  # in bytes
    num_files_remaining = len(files)
    for f in files:
        fsize = os.stat(f).st_size
        if total_size < max_total_size and num_files_remaining <= num_files_keep:
            break
        total_size -= fsize
        num_files_remaining -= 1
        try:
            os.remove(f)
        except OSError as e:
            _logger.warning(f"cannot delete old logfile: {e}")


_logfile_path = None
def _configure_file_logging(
    log_directory: pathlib.Path,
    *,
    num_files_keep: int,
    max_total_size: int,
):
    from .util import os_chmod

    global _logfile_path
    assert _logfile_path is None, 'file logging already initialized'
    log_directory.mkdir(exist_ok=True, mode=0o700)

    _delete_old_logs(log_directory, num_files_keep=num_files_keep, max_total_size=max_total_size)

    timestamp = datetime.datetime.now(datetime.timezone.utc).strftime("%Y%m%dT%H%M%SZ")
    PID = os.getpid()
    _logfile_path = log_directory / f"electrum_log_{timestamp}_{PID}.log"
    # we create the file with restrictive perms, instead of letting FileHandler create it
    with open(_logfile_path, "w+") as f:
        os_chmod(_logfile_path, 0o600)

    logfile_backupcount = 4
    file_handler = logging.handlers.RotatingFileHandler(
        _logfile_path,
        maxBytes=max_total_size // (logfile_backupcount+1),
        backupCount=logfile_backupcount,
        encoding='utf-8')
    file_handler.setFormatter(file_formatter)
    file_handler.setLevel(logging.DEBUG)
    root_logger.addHandler(file_handler)
    if _inmemory_startup_logs:
        _inmemory_startup_logs.dump_to_target(file_handler)


console_stderr_handler = None
def _configure_stderr_logging(*, verbosity=None):
    # log to stderr; by default only WARNING and higher
    global console_stderr_handler
    if console_stderr_handler is not None:
        _logger.warning("stderr handler already exists")
        return
    console_stderr_handler = logging.StreamHandler(sys.stderr)
    console_stderr_handler.setFormatter(console_formatter)
    if not verbosity:
        console_stderr_handler.setLevel(logging.WARNING)
        root_logger.addHandler(console_stderr_handler)
    else:
        console_stderr_handler.setLevel(logging.DEBUG)
        root_logger.addHandler(console_stderr_handler)
        _process_verbosity_log_levels(verbosity)
    if _inmemory_startup_logs:
        _inmemory_startup_logs.dump_to_target(console_stderr_handler)


def _process_verbosity_log_levels(verbosity):
    if verbosity == '*' or not isinstance(verbosity, str):
        return
    # example verbosity:
    #   debug,network=error,interface=error      // effectively blacklists network and interface
    #   warning,network=debug,interface=debug    // effectively whitelists network and interface
    filters = verbosity.split(',')
    for filt in filters:
        if not filt: continue
        items = filt.split('=')
        if len(items) == 1:
            level = items[0]
            electrum_logger.setLevel(level.upper())
        elif len(items) == 2:
            logger_name, level = items
            logger = get_logger(logger_name)
            logger.setLevel(level.upper())
        else:
            raise Exception(f"invalid log filter: {filt}")


class _CustomLogger(logging.getLoggerClass()):
    def __init__(self, name, *args, **kwargs):
        super().__init__(name, *args, **kwargs)
        self.msg_hashes_seen = set()  # type: Set[bytes]
        # ^ note: size grows without bounds, but only for log lines using "only_once".

    def _log(self, level, msg: str, *args, only_once: bool = False, **kwargs) -> None:
        """Overridden to add 'only_once' arg to logger.debug()/logger.info()/logger.warning()/etc."""
        if only_once:  # if set, this logger will only log this msg a single time during its lifecycle
            msg_hash = hashlib.sha256(msg.encode("utf-8")).digest()
            if msg_hash in self.msg_hashes_seen:
                return
            self.msg_hashes_seen.add(msg_hash)
        super()._log(level, msg, *args, **kwargs)

logging.setLoggerClass(_CustomLogger)


# enable logs universally (including for other libraries)
root_logger = logging.getLogger()
root_logger.setLevel(logging.WARNING)

# Start collecting log messages now, into an in-memory buffer. This buffer is only
# used until the proper log handlers are fully configured, including their verbosity,
# at which point we will dump its contents into those, and remove this log handler.
# Note: this is set up at import-time instead of e.g. as part of a function that is
#       called from run_electrum (the main script). This is to have this run as early
#       as possible.
# Note: some users might use Electrum as a python library and not use run_electrum,
#       in which case these logs might never get redirected or cleaned up.
#       Also, the python docs recommend libraries not to set a handler, to
#       avoid interfering with the user's logging.
_inmemory_startup_logs = None
if getattr(sys, "_ELECTRUM_RUNNING_VIA_RUNELECTRUM", False):
    _inmemory_startup_logs = TruncatingMemoryHandler()
    root_logger.addHandler(_inmemory_startup_logs)

# creates a logger specifically for electrum library
electrum_logger = logging.getLogger("electrum")
electrum_logger.setLevel(logging.DEBUG)


# --- External API

def get_logger(name: str) -> _CustomLogger:
    prefix = "electrum."
    if name.startswith(prefix):
        name = name[len(prefix):]
    return electrum_logger.getChild(name)


_logger = get_logger(__name__)
_logger.setLevel(logging.INFO)


class Logger:

    def __init__(self):
        self.logger = self.__get_logger_for_obj()

    def __get_logger_for_obj(self) -> logging.Logger:
        cls = self.__class__
        if cls.__module__:
            name = f"{cls.__module__}.{cls.__name__}"
        else:
            name = cls.__name__
        try:
            diag_name = self.diagnostic_name()
        except Exception as e:
            raise Exception("diagnostic name not yet available?") from e
        if diag_name:
            name += f".[{diag_name}]"
        logger = get_logger(name)
        return logger

    def diagnostic_name(self):
        return ''


def configure_logging(config: 'SimpleConfig', *, log_to_file: Optional[bool] = None) -> None:
    from .util import is_android_debug_apk

    verbosity = config.get('verbosity')
    if not verbosity and config.GUI_ENABLE_DEBUG_LOGS:
        verbosity = '*'
    _configure_stderr_logging(verbosity=verbosity)

    if log_to_file is None:
        log_to_file = config.WRITE_LOGS_TO_DISK
        log_to_file |= is_android_debug_apk()
    if log_to_file:
        log_directory = pathlib.Path(config.path) / "logs"
        num_files_keep = config.LOGS_NUM_FILES_KEEP
        max_total_size = config.LOGS_MAX_TOTAL_SIZE_BYTES
        _configure_file_logging(log_directory, num_files_keep=num_files_keep, max_total_size=max_total_size)

    # clean up and delete in-memory logs
    global _inmemory_startup_logs
    if _inmemory_startup_logs:
        num_discarded = _inmemory_startup_logs.num_messages_seen - _inmemory_startup_logs.max_size
        if num_discarded > 0:
            _logger.warning(f"Too many log messages! Some have been discarded. "
                            f"(discarded {num_discarded} messages)")
        _inmemory_startup_logs.close()
        root_logger.removeHandler(_inmemory_startup_logs)
        _inmemory_startup_logs = None

    from . import ELECTRUM_VERSION
    from .constants import GIT_REPO_URL
    _logger.info(f"Electrum version: {ELECTRUM_VERSION} - https://electrum.org - {GIT_REPO_URL}")
    _logger.info(f"Python version: {sys.version}. On platform: {describe_os_version()}")
    _logger.info(f"Logging to file: {str(_logfile_path)}")
    _logger.info(f"Log filters: verbosity {repr(verbosity)}")


def get_logfile_path() -> Optional[pathlib.Path]:
    return _logfile_path


def describe_os_version() -> str:
    if 'ANDROID_DATA' in os.environ:
        import jnius
        bv = jnius.autoclass('android.os.Build$VERSION')
        b = jnius.autoclass('android.os.Build')
        return "Android {} on {} {} ({})".format(bv.RELEASE, b.BRAND, b.DEVICE, b.DISPLAY)
    else:
        return platform.platform()


def get_git_version() -> Optional[str]:
    dir = os.path.dirname(os.path.realpath(__file__))
    try:
        version = subprocess.check_output(
            ['git', 'describe', '--always', '--dirty'], cwd=dir)
        version = str(version, "utf8").strip()
    except Exception:
        version = None
    return version



#!/usr/bin/env python3
#
# Electron Cash - lightweight Bitcoin client
# Copyright (C) 2019 Axel Gembe <derago@gmail.com>
#
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction,
# including without limitation the rights to use, copy, modify, merge,
# publish, distribute, sublicense, and/or sell copies of the Software,
# and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import sys
import ctypes
import os
from typing import List
from enum import IntEnum

from ..logging import get_logger

from . import MissingLib
from .abstract_base import AbstractQrCodeReader, QrCodeResult


_logger = get_logger(__name__)

if 'ANDROID_DATA' in os.environ:
    LIBNAME = 'libzbar.so'
elif sys.platform == 'darwin':
    LIBNAME = 'libzbar.0.dylib'
elif sys.platform in ('windows', 'win32'):
    LIBNAME = 'libzbar-0.dll'
else:
    LIBNAME = 'libzbar.so.0'

try:
    try:
        LIBZBAR = ctypes.cdll.LoadLibrary(os.path.join(os.path.dirname(__file__), '..', LIBNAME))
    except OSError as e:
        LIBZBAR = ctypes.cdll.LoadLibrary(LIBNAME)

    LIBZBAR.zbar_image_create.restype = ctypes.c_void_p
    LIBZBAR.zbar_image_scanner_create.restype = ctypes.c_void_p
    LIBZBAR.zbar_image_scanner_get_results.restype = ctypes.c_void_p
    LIBZBAR.zbar_symbol_set_first_symbol.restype = ctypes.c_void_p
    LIBZBAR.zbar_symbol_get_data.restype = ctypes.POINTER(ctypes.c_char_p)
    LIBZBAR.zbar_image_scanner_set_config.argtypes = [ctypes.c_void_p, ctypes.c_int, ctypes.c_int, ctypes.c_int]
    LIBZBAR.zbar_image_set_sequence.argtypes = [ctypes.c_void_p, ctypes.c_int]
    LIBZBAR.zbar_image_set_size.argtypes = [ctypes.c_void_p, ctypes.c_int, ctypes.c_int]
    LIBZBAR.zbar_image_set_format.argtypes = [ctypes.c_void_p, ctypes.c_int]
    LIBZBAR.zbar_image_set_data.argtypes = [ctypes.c_void_p, ctypes.c_void_p, ctypes.c_int, ctypes.c_void_p]
    LIBZBAR.zbar_image_scanner_recycle_image.argtypes = [ctypes.c_void_p, ctypes.c_void_p]
    LIBZBAR.zbar_scan_image.argtypes = [ctypes.c_void_p, ctypes.c_void_p]
    LIBZBAR.zbar_image_scanner_get_results.argtypes = [ctypes.c_void_p]
    LIBZBAR.zbar_symbol_set_first_symbol.argtypes = [ctypes.c_void_p]
    LIBZBAR.zbar_symbol_get_data_length.argtypes = [ctypes.c_void_p]
    LIBZBAR.zbar_symbol_get_data.argtypes = [ctypes.c_void_p]
    LIBZBAR.zbar_symbol_get_loc_size.argtypes = [ctypes.c_void_p]
    LIBZBAR.zbar_symbol_get_loc_x.argtypes = [ctypes.c_void_p, ctypes.c_int]
    LIBZBAR.zbar_symbol_get_loc_y.argtypes = [ctypes.c_void_p, ctypes.c_int]
    LIBZBAR.zbar_symbol_next.argtypes = [ctypes.c_void_p]
    LIBZBAR.zbar_image_scanner_destroy.argtypes = [ctypes.c_void_p]
    LIBZBAR.zbar_image_destroy.argtypes = [ctypes.c_void_p]

    #if is_verbose:
        #LIBZBAR.zbar_set_verbosity(100)
except OSError:
    _logger.exception("Failed to load zbar")
    LIBZBAR = None

FOURCC_Y800 = 0x30303859

@ctypes.CFUNCTYPE(None, ctypes.c_void_p)
def zbar_cleanup(image):
    """
    Do nothing, this is just so zbar doesn't try to manage our QImage buffers
    """

class ZbarSymbolType(IntEnum):
    """
    Supported symbol types, see zbar_symbol_type_e in zbar.h
    """
    EAN2 = 2
    EAN5 = 5
    EAN8 = 8
    UPCE = 9
    ISBN10 = 10
    UPCA = 12
    EAN13 = 13
    ISBN13 = 14
    COMPOSITE = 15
    I25 = 25
    DATABAR = 34
    DATABAR_EXP = 35
    CODABAR = 38
    CODE39 = 39
    PDF417 = 57
    QRCODE = 64
    SQCODE = 80
    CODE93 = 93
    CODE128 = 128

class ZbarConfig(IntEnum):
    """
    Supported configuration options, see zbar_config_e in zbar.h
    """
    ENABLE = 0

class ZbarQrCodeReader(AbstractQrCodeReader):
    """
    Reader that uses libzbar
    """

    def __init__(self):
        if not LIBZBAR:
            raise MissingLib('Zbar library not found')
        # Set up zbar
        self.zbar_scanner = LIBZBAR.zbar_image_scanner_create()
        self.zbar_image = LIBZBAR.zbar_image_create()

        # Disable all symbols
        for sym_type in ZbarSymbolType:
            LIBZBAR.zbar_image_scanner_set_config(self.zbar_scanner, sym_type, ZbarConfig.ENABLE, 0)

        # Enable only QR codes
        LIBZBAR.zbar_image_scanner_set_config(self.zbar_scanner, ZbarSymbolType.QRCODE,
                                              ZbarConfig.ENABLE, 1)

    def __del__(self):
        if LIBZBAR:
            LIBZBAR.zbar_image_scanner_destroy(self.zbar_scanner)
            LIBZBAR.zbar_image_destroy(self.zbar_image)

    def read_qr_code(self, buffer: ctypes.c_void_p, buffer_size: int,
                     rowlen_bytes: int,  # this param is ignored in this implementation
                     width: int, height: int, frame_id: int = -1) -> List[QrCodeResult]:
        LIBZBAR.zbar_image_set_sequence(self.zbar_image, frame_id)
        LIBZBAR.zbar_image_set_size(self.zbar_image, width, height)
        LIBZBAR.zbar_image_set_format(self.zbar_image, FOURCC_Y800)
        LIBZBAR.zbar_image_set_data(self.zbar_image, buffer, buffer_size, zbar_cleanup)
        LIBZBAR.zbar_image_scanner_recycle_image(self.zbar_scanner, self.zbar_image)
        LIBZBAR.zbar_scan_image(self.zbar_scanner, self.zbar_image)

        result_set = LIBZBAR.zbar_image_scanner_get_results(self.zbar_scanner)

        res = []
        symbol = LIBZBAR.zbar_symbol_set_first_symbol(result_set)
        while symbol:
            symbol_data_len = LIBZBAR.zbar_symbol_get_data_length(symbol)
            symbol_data_ptr = LIBZBAR.zbar_symbol_get_data(symbol)
            symbol_data_bytes = ctypes.string_at(symbol_data_ptr, symbol_data_len)
            symbol_data = symbol_data_bytes.decode('utf-8')

            symbol_loc = []
            symbol_loc_len = LIBZBAR.zbar_symbol_get_loc_size(symbol)
            for i in range(0, symbol_loc_len):
                # Normalize the coordinates into 0..1 range by dividing by width / height
                symbol_loc_x = LIBZBAR.zbar_symbol_get_loc_x(symbol, i)
                symbol_loc_y = LIBZBAR.zbar_symbol_get_loc_y(symbol, i)
                symbol_loc.append((symbol_loc_x, symbol_loc_y))

            # Find the center by getting the average values of the corners x and y coordinates
            symbol_loc_sum_x = sum([l[0] for l in symbol_loc])
            symbol_loc_sum_y = sum([l[1] for l in symbol_loc])
            symbol_loc_center = (int(symbol_loc_sum_x / symbol_loc_len), int(symbol_loc_sum_y / symbol_loc_len))

            res.append(QrCodeResult(symbol_data, symbol_loc_center, symbol_loc))

            symbol = LIBZBAR.zbar_symbol_next(symbol)

        return res



from datetime import datetime, timedelta
from decimal import Decimal

from PyQt6.QtCore import pyqtProperty, pyqtSignal, pyqtSlot, QObject, QRegularExpression

from electrum.bitcoin import COIN
from electrum.exchange_rate import FxThread
from electrum.logging import get_logger
from electrum.simple_config import SimpleConfig
from electrum.util import event_listener

from electrum.gui.common_qt.util import QtEventListener

from .qetypes import QEAmount


class QEFX(QObject, QtEventListener):
    _logger = get_logger(__name__)

    quotesUpdated = pyqtSignal()

    def __init__(self, fxthread: FxThread, config: SimpleConfig, parent=None):
        super().__init__(parent)
        self.fx = fxthread
        self.config = config
        self.register_callbacks()
        self.destroyed.connect(lambda: self.on_destroy())

    def on_destroy(self):
        self.unregister_callbacks()

    @event_listener
    def on_event_on_quotes(self, *args):
        self._logger.debug('new quotes')
        self.quotesUpdated.emit()

    historyUpdated = pyqtSignal()
    @event_listener
    def on_event_on_history(self, *args):
        self._logger.debug('new history')
        self.historyUpdated.emit()

    currenciesChanged = pyqtSignal()
    @pyqtProperty('QVariantList', notify=currenciesChanged)
    def currencies(self):
        return self.fx.get_currencies(self.historicRates)

    rateSourcesChanged = pyqtSignal()
    @pyqtProperty('QVariantList', notify=rateSourcesChanged)
    def rateSources(self):
        return self.fx.get_exchanges_by_ccy(self.fiatCurrency, self.historicRates)

    fiatCurrencyChanged = pyqtSignal()
    @pyqtProperty(str, notify=fiatCurrencyChanged)
    def fiatCurrency(self):
        return self.fx.get_currency()

    @fiatCurrency.setter
    def fiatCurrency(self, currency):
        if currency != self.fiatCurrency:
            self.fx.set_currency(currency)
            self.enabled = self.enabled and currency != ''
            self.fiatCurrencyChanged.emit()
            self.rateSourcesChanged.emit()

    @pyqtProperty('QRegularExpression', notify=fiatCurrencyChanged)
    def fiatAmountRegex(self):
        decimals = self.fx.ccy_precision()
        exp = '[0-9]*'
        if decimals:
            exp += '\\.'
            exp += '[0-9]{0,%d}' % decimals
        return QRegularExpression(exp)

    historicRatesChanged = pyqtSignal()
    @pyqtProperty(bool, notify=historicRatesChanged)
    def historicRates(self):
        if not self.fx.config.cv.FX_HISTORY_RATES.is_set():
            self.fx.config.FX_HISTORY_RATES = True  # override default
        return self.fx.config.FX_HISTORY_RATES

    @historicRates.setter
    def historicRates(self, checked):
        if checked != self.historicRates:
            self.fx.config.FX_HISTORY_RATES = bool(checked)
            self.historicRatesChanged.emit()
            self.rateSourcesChanged.emit()

    rateSourceChanged = pyqtSignal()
    @pyqtProperty(str, notify=rateSourceChanged)
    def rateSource(self):
        return self.fx.config_exchange()

    @rateSource.setter
    def rateSource(self, source):
        if source != self.rateSource:
            self.fx.set_exchange(source)
            self.rateSourceChanged.emit()

    enabledUpdated = pyqtSignal()  # curiously, enabledChanged is clashing, so name it enabledUpdated
    @pyqtProperty(bool, notify=enabledUpdated)
    def enabled(self):
        return self.fx.is_enabled()

    @enabled.setter
    def enabled(self, enable):
        if enable != self.enabled:
            self.fx.set_enabled(enable)
            self.enabledUpdated.emit()

    @pyqtSlot(str, result=str)
    @pyqtSlot(str, bool, result=str)
    @pyqtSlot(QEAmount, result=str)
    @pyqtSlot(QEAmount, bool, result=str)
    def fiatValue(self, satoshis, plain=True):
        rate = self.fx.exchange_rate()
        if isinstance(satoshis, QEAmount):
            satoshis = satoshis.msatsInt / 1000 if satoshis.msatsInt != 0 else satoshis.satsInt
        else:
            try:
                sd = Decimal(satoshis)
            except Exception:
                return ''
        if plain:
            return self.fx.ccy_amount_str(self.fx.fiat_value(satoshis, rate), add_thousands_sep=False)
        else:
            return self.fx.value_str(satoshis, rate)

    @pyqtSlot(str, str, result=str)
    @pyqtSlot(str, str, bool, result=str)
    @pyqtSlot(QEAmount, str, result=str)
    @pyqtSlot(QEAmount, str, bool, result=str)
    def fiatValueHistoric(self, satoshis, timestamp, plain=True):
        if isinstance(satoshis, QEAmount):
            satoshis = satoshis.msatsInt / 1000 if satoshis.msatsInt != 0 else satoshis.satsInt
        else:
            try:
                sd = Decimal(satoshis)
            except Exception:
                return ''

        try:
            td = Decimal(timestamp)
            if td == 0:
                return ''
        except Exception:
            return ''
        dt = datetime.fromtimestamp(int(td))
        if plain:
            return self.fx.ccy_amount_str(self.fx.historical_value(satoshis, dt), add_thousands_sep=False)
        else:
            return self.fx.historical_value_str(satoshis, dt)

    @pyqtSlot(str, result=str)
    @pyqtSlot(str, bool, result=str)
    def satoshiValue(self, fiat, plain=True):
        rate = self.fx.exchange_rate()
        try:
            fd = Decimal(fiat)
        except Exception:
            return ''
        v = fd / Decimal(rate) * COIN
        if v.is_nan():
            return ''
        if plain:
            return str(v.to_integral_value())
        else:
            return self.config.format_amount(v)

    @pyqtSlot(str, result=bool)
    def isRecent(self, timestamp):
        # return True if unknown, e.g. timestamp not known yet, tx in mempool
        try:
            td = Decimal(timestamp)
            if td == 0:
                return True
        except Exception:
            return True
        dt = datetime.fromtimestamp(int(td))
        return dt + timedelta(days=1) > datetime.today()



#!/usr/bin/env python3
import datetime
import glob
import os
import re
import subprocess
import sys
import io
import zipfile

# check dependencies are available
try:
    import requests
except ImportError as e:
    sys.exit(f"Error: {str(e)}. Try 'python3 -m pip install --user <module-name>' (or 'python3-requests' from Debian)")

try:
    import polib
except ImportError as e:
    sys.exit(f"Error: {str(e)}. Try 'python3 -m pip install --user <module-name>' (or 'python3-polib' from Debian)")

try:
    subprocess.check_output(["msgattrib", "--version"])
except (subprocess.CalledProcessError, OSError):
    raise Exception("missing gettext. Maybe try 'apt install gettext'")


crowdin_project_id = 20482  # for "Electrum" project on crowdin

BITCOIN_ADDRESS_REGEXP = re.compile('([13]|bc1)[a-zA-Z0-9]{30,}')
EMAIL_ADDRESS_REGEXP = re.compile(r'[\w.+-]+@[\w-]+\.[\w.-]+')

URL1_REGEXP = re.compile(r"\S+\.\S*\w+\S*/\S+")  # str has a "." and somewhere later has a "/" ==> URL?
assert URL1_REGEXP.search("download security update from scamwebsite.com/click-here") is not None
URL2_REGEXP = re.compile(r"http(s){0,1}://")

MIXED_LETTERS_AND_DIGITS_WORD_REGEXP = re.compile(  # try to match any cryptocurrency address
    r"(?a)"  # limit match to ascii, as CJK languages do not put whitespaces between words (would match full sentence otherwise)
    r"(?=\w{16,})"  # positive lookahead: check word length >=16
    r"("
        r"(\w*[a-zA-Z]\w*[0-9]\w*)|"  # contains a letter and then later a digit, OR
        r"(\w*[0-9]\w*[a-zA-Z]\w*)"   # contains a digit and then later a letter
    r")"
)
assert not (MIXED_LETTERS_AND_DIGITS_WORD_REGEXP.search("bip39 seeds cannot be converted to electrum seeds") is not None)
assert not (MIXED_LETTERS_AND_DIGITS_WORD_REGEXP.search("ライトニングは現在p2wpkhアドレスのHDウォレットでのみ利用可能です。") is not None)
assert MIXED_LETTERS_AND_DIGITS_WORD_REGEXP.search("ラ TB1Q9DVVAG8NS2EPXZ285FW03CR78HWEJP0Z7DFAFE") is not None
assert MIXED_LETTERS_AND_DIGITS_WORD_REGEXP.search("bitcoin:tb1q9dvvag8ns2epxz285fw03cr78hwejp0z7dfafe?amount=12345678") is not None
assert MIXED_LETTERS_AND_DIGITS_WORD_REGEXP.search("0x456d9347342B72BCf800bBf117391ac2f807c6bF") is not None  # eth
assert MIXED_LETTERS_AND_DIGITS_WORD_REGEXP.search("84EgZVjXKF4d1JkEhZSxm4LQQEx64AvqQEwkvWPtHEb5JMrB1Y86y1vCPSCiXsKzbfS9x8vCpx3gVgPaHCpobPYqQzANTnC") is not None  # xmr


def get_crowdin_api_key() -> str:
    crowdin_api_key = None
    if "crowdin_api_key" in os.environ:
        return os.environ["crowdin_api_key"]
    filename = os.path.expanduser('~/.crowdin_api_key')
    if os.path.exists(filename):
        with open(filename) as f:
            crowdin_api_key = f.read().strip()
    return crowdin_api_key


def pull_locale(path, *, crowdin_api_key=None):
    global_headers = {}
    if crowdin_api_key is None:
        crowdin_api_key = get_crowdin_api_key()
    if not crowdin_api_key:
        # Looks like crowdin does not even allow downloading without auth anymore.
        raise Exception("missing required crowdin_api_key")
    if crowdin_api_key:
        global_headers["Authorization"] = "Bearer {}".format(crowdin_api_key)

    if not os.path.exists(path):
        os.mkdir(path)
    os.chdir(path)

    # note: We won't request a build now, instead we download the latest build.
    #       This assumes that the push_locale script was run recently (in the past few days).
    print('Getting list of builds from crowdin...')
    # https://support.crowdin.com/developer/api/v2/?q=api#tag/Translations/operation/api.projects.translations.builds.getMany
    url = f'https://api.crowdin.com/api/v2/projects/{crowdin_project_id}/translations/builds'
    headers = {**global_headers, **{"content-type": "application/json"}}
    response = requests.request("GET", url, headers=headers)
    response.raise_for_status()
    print("", "translations.builds.getMany:", "-" * 20, response.text, "-" * 20, sep="\n")

    latest_build = response.json()["data"][0]["data"]
    assert latest_build["status"] == "finished", latest_build["status"]
    # if latest_build["attributes"]["exportApprovedOnly"] is not True:
    #     raise Exception("latest_build from crowdin MUST have exportApprovedOnly==true")
    created_at = datetime.datetime.fromisoformat(latest_build["createdAt"])
    if (datetime.datetime.now(datetime.timezone.utc) - created_at) > datetime.timedelta(days=2):
        raise Exception(f"latest translation build looks too old. {created_at.isoformat()=}")
    build_id = latest_build["id"]

    print('Asking crowdin to generate a URL for the latest build...')
    # https://support.crowdin.com/developer/api/v2/?q=api#tag/Translations/operation/api.projects.translations.builds.download.download
    url = f'https://api.crowdin.com/api/v2/projects/{crowdin_project_id}/translations/builds/{build_id}/download'
    headers = {**global_headers, **{"content-type": "application/json"}}
    response = requests.request("GET", url, headers=headers)
    response.raise_for_status()
    print("", "translations.builds.download.download:", "-" * 20, response.text, "-" * 20, sep="\n")

    build_url = response.json()["data"]["url"]

    # Download & unzip
    print('Downloading translations...')
    response = requests.request('GET', build_url, headers={})
    response.raise_for_status()
    s = response.content
    zfobj = zipfile.ZipFile(io.BytesIO(s))

    print('Unzipping translations...')
    prefix = "electrum-client/locale/"
    for name in zfobj.namelist():
        if not name.startswith(prefix) or name == prefix:
            continue
        if name.endswith('/'):
            if not os.path.exists(name[len(prefix):]):
                os.mkdir(name[len(prefix):])
        else:
            name_suffix = name[len(prefix):]
            with open(name_suffix, 'wb') as output:
                output.write(zfobj.read(name))
            if name.endswith('.po'):
                filter_exclude_comment_lines(name_suffix)
                filter_exclude_untranslated_strings(name_suffix)
            else:
                raise Exception(f"unexpected file inside zipfile from crowdin: {name}")


def filter_exclude_comment_lines(fname: str):
    """Remove lines starting with "#". ==> easier to review diffs

    As auto-generated comments contain line numbers, removing them makes
    reviewing diffs much more practical.

    note: we could be more relaxed and only rm lines starting with "#:",
          see https://www.gnu.org/software/gettext/manual/html_node/PO-File-Entries.html
    """
    os.rename(fname, f"{fname}.orig")
    with open(f"{fname}.orig", "r+", encoding="utf-8") as f_orig:
        with open(fname, "w", encoding="utf-8") as f_filtered:
            for line in f_orig:
                if not line.startswith("#"):
                    f_filtered.write(line)
    os.remove(f"{fname}.orig")


def filter_exclude_untranslated_strings(fname: str):
    """Remove strings with empty translations. ==> easier to review diffs"""
    cmd = ["msgattrib", "--translated", "--no-wrap", "--output-file", fname, fname]
    subprocess.check_output(cmd)


def detect_malicious_stuff_in_dir(path_locale: str) -> None:
    is_detected = False
    # scan each .po file separately
    files_list = glob.glob(f"{path_locale}/**/*.po", recursive=True)
    for fname in files_list:
        is_detected |= detect_malicious_stuff_in_po_file(fname)
    # after finding all errors, exit now if there were any:
    if is_detected:
        raise Exception("detected some possibly malicious translations. see logs above.")


def detect_malicious_stuff_in_po_file(fname: str) -> bool:
    pofile = polib.pofile(fname)

    is_detected = False
    regexes = {
        "BITCOIN_ADDRESS_REGEXP": BITCOIN_ADDRESS_REGEXP,
        "EMAIL_ADDRESS_REGEXP": EMAIL_ADDRESS_REGEXP,
        "URL1_REGEXP": URL1_REGEXP,
        "URL2_REGEXP": URL2_REGEXP,
        "MIXED_LETTERS_AND_DIGITS_WORD_REGEXP": MIXED_LETTERS_AND_DIGITS_WORD_REGEXP,
    }
    for entry in pofile:
        for regex_name, regex in regexes.items():
            if regex.search(entry.msgstr) is not None:
                print(
                    f">> regex {regex_name} matched in {fname!r},\n"
                    f"\t{entry.msgid=!r}\n"
                    f"\t{entry.msgstr=!r}\n")
                is_detected = True

    return is_detected


if __name__ == '__main__':
    path_here = os.path.dirname(os.path.realpath(__file__))
    path_locale = os.path.join(path_here, "locale")

    pull_locale(path_locale)
    detect_malicious_stuff_in_dir(path_locale)

    print('Local updates done.')
    print("Please don't commit directly to master, switch to a branch instead.")
    c = input("Do you want to git commit this? (y/n): ")
    if c != "y":
        sys.exit(0)

    print('Preparing git commit...')
    os.chdir(path_here)
    for lang in os.listdir('locale'):
        po = 'locale/%s/electrum.po' % lang
        cmd = "git add %s"%po
        os.system(cmd)

    os.system("git commit -a -m 'update translations'")
    print("please push to a branch, and open a pull request")



import sys
from typing import TYPE_CHECKING, Optional

from PyQt6.QtCore import pyqtSignal, pyqtProperty, QObject

from electrum.logging import get_logger

if TYPE_CHECKING:
    from electrum.gui.qml import ElectrumQmlApplication
    from electrum.plugin import BasePlugin


class PluginQObject(QObject):
    logger = get_logger(__name__)

    pluginChanged = pyqtSignal()
    busyChanged = pyqtSignal()
    pluginEnabledChanged = pyqtSignal()

    def __init__(self, plugin: 'BasePlugin', parent: Optional['ElectrumQmlApplication']):
        super().__init__(parent)

        self._busy = False

        self.plugin = plugin
        self.app = parent

    @pyqtProperty(str, notify=pluginChanged)
    def name(self): return self._name

    @pyqtProperty(bool, notify=busyChanged)
    def busy(self): return self._busy

    # below only used for QML, not compatible yet with Qt

    @pyqtProperty(bool, notify=pluginEnabledChanged)
    def pluginEnabled(self): return self.plugin.is_enabled()

    @pluginEnabled.setter
    def pluginEnabled(self, enabled):
        if enabled != self.plugin.is_enabled():
            self.logger.debug(f'can {self.plugin.can_user_disable()}, {self.plugin.is_available()}')
            if not self.plugin.can_user_disable() and not enabled:
                return
            if enabled:
                self.app.plugins.enable(self.plugin.name)
            else:
                self.app.plugins.disable(self.plugin.name)
            self.pluginEnabledChanged.emit()

